@@ -14,7 +14,8 @@ import  textwrap
 import  contextlib
 import  subprocess
 import  pickle
-from    datetime            import  datetime
+import  shutil
+from    datetime            import  datetime, timedelta
 from    concurrent.futures  import  ProcessPoolExecutor
 
 import  numpy               as      np
@@ -43,7 +44,7 @@ signal_handler = SignalHandlerKI()
 
 
 
-@autorepr(short = {"script", "scheduler_cmd"})
+@autorepr(short = {"script"})
 class AccessSetup:
     '''Structure storing constant attributes for an ACCES optimisation run.
 
@@ -65,10 +66,6 @@ class AccessSetup:
     script : str
         The modified user script that will be executed.
 
-    scheduler_cmd : list
-        The list of commands the will be prepended to each script execution,
-        defined by the `Scheduler`.
-
     population : int
         The number of simulations to be run in parallel in each epoch.
 
@@ -83,7 +80,7 @@ class AccessSetup:
         The random number generator used, seeded with `seed`.
     '''
 
-    def __init__(self, script_path, scheduler):
+    def __init__(self, script_path):
         '''Given a path to a user-defined simulation script, extract the free
         parameters and generate the ACCES script.
         '''
@@ -93,15 +90,6 @@ class AccessSetup:
         self.rng = None
         self.seed = None
 
-        # Type-check and generate scheduler commands
-        if not isinstance(scheduler, schedulers.Scheduler):
-            raise TypeError(textwrap.fill((
-                "The input `scheduler` must be a subclass of `coexist."
-                f"schedulers.Scheduler`. Received {type(scheduler)}."
-            )))
-
-        self.scheduler_cmd = scheduler.generate()
-
         # Extract parameters and generate ACCES script
         with open(script_path, "r") as f:
             user_code = f.readlines()
@@ -514,7 +502,12 @@ class AccessPaths:
         for attr in ["results", "outputs", "script", "setup", "epochs",
                      "epochs_scaled", "history", "history_scaled"]:
             prev = getattr(self, attr)
-            setattr(self, attr, os.path.join(prefix, os.path.split(prev)[1]))
+            if isinstance(prev, str):
+                setattr(
+                    self,
+                    attr,
+                    os.path.join(prefix, os.path.split(prev)[1])
+                )
 
 
     def save_history(self, setup, progress):
@@ -613,6 +606,22 @@ class AccessPaths:
             )
 
 
+    def copy(self):
+        '''Create a copy of an `AccessPaths` object.
+        '''
+
+        return AccessPaths(
+            directory = self.directory,
+            results = self.results,
+            outputs = self.outputs,
+            script = self.script,
+            setup = self.setup,
+            epochs = self.epochs,
+            epochs_scaled = self.epochs_scaled,
+            history = self.history,
+            history_scaled = self.history_scaled,
+        )
+
 
 
 @autorepr(short = True)
@@ -737,8 +746,8 @@ class AccessProgress:
         deviation, errors and the combined total error.
         '''
         results = []
-        stdout_rec = []
-        stderr_rec = []
+        # stdout_rec = []
+        # stderr_rec = []
         crashed = []
 
         # Occasionally check if jobs finished
@@ -754,31 +763,7 @@ class AccessProgress:
                 wait = 0
                 for i, proc in enumerate(processes):
                     proc_index = int(proc.args[-1].split(".")[-2])
-                    stdout, stderr = proc.communicate()
-
-                    # If a *new* output message was recorded in stdout, log it
-                    if len(stdout) and stdout != self.stdout:
-                        stdout_rec.append(proc_index)
-                        self.stdout = stdout
-
-                        stdout_log = os.path.join(
-                            paths.outputs,
-                            f"stdout.{proc_index}.log",
-                        )
-                        with open(stdout_log, "w") as f:
-                            f.write(self.stdout)
-
-                    # If a *new* error message was recorded in stderr, log it
-                    if len(stderr) and stderr != self.stderr:
-                        stderr_rec.append(proc_index)
-                        self.stderr = stderr
-
-                        stderr_log = os.path.join(
-                            paths.outputs,
-                            f"stderr.{proc_index}.log",
-                        )
-                        with open(stderr_log, "w") as f:
-                            f.write(self.stderr)
+                    proc.communicate()
 
                     # Load result if the file exists, otherwise set it to NaN
                     if os.path.isfile(result_paths[i]):
@@ -822,7 +807,7 @@ class AccessProgress:
             waited += wait
             wait = min(wait * 1.5, 60)
 
-        return results, stdout_rec, stderr_rec, crashed
+        return results, crashed
 
 
 
@@ -898,14 +883,11 @@ class Access:
 
     Attributes
     ----------
-    script : str
-        The ACCES code generated from the user-supplied script.
-
     setup : coexist.access.AccessSetup
         Structure storing given ACCES configuration, containing the free
-        ``parameters``, scheduler commands ``scheduler_cmd``, number of
-        solutions to try in parallel ``population``, target uncertainty
-        ``target``, seeded random number generator ``rng`` and the ``seed``.
+        ``parameters``, number of solutions to try in parallel ``population``,
+        target uncertainty ``target``, seeded random number generator ``rng``
+        and the ``seed``.
 
     paths : coexist.access.AccessPaths
         Structure storing paths to the ACCES ``directory``, saved ``state``,
@@ -917,7 +899,14 @@ class Access:
         ``history`` (and CMA-scaled versions of them) and latest ``stdout`` and
         ``stderr`` messages.
 
-    verbose : int
+    scheduler : coexist.schedulers.Scheduler subclass
+        The scheduler used to launch each simulation in parallel.
+
+    multi_objective : object, optional
+        An object defining the method `combine`, combining multiple errors into
+        a single value.
+
+    verbose : int, optional
         Integer denoting the level of verbosity, where 0 is quiet and 5 is
         maximally verbose.
     '''
@@ -948,9 +937,19 @@ class Access:
         '''
 
         # Creating class attributes
-        self.setup = AccessSetup(script_path, scheduler)
+        self.setup = AccessSetup(script_path)
         self.paths = AccessPaths()
         self.progress = AccessProgress()
+
+        # Type-check scheduler
+        if not isinstance(scheduler, schedulers.Scheduler):
+            raise TypeError(textwrap.fill((
+                "The input `scheduler` must be a subclass of `coexist."
+                f"schedulers.Scheduler`. Received {type(scheduler)}."
+            )))
+        self.scheduler = scheduler
+
+        # Will be set in `learn`
         self.multi_objective = None
         self.verbose = None
 
@@ -1144,9 +1143,31 @@ class Access:
 
         head = "=" * 80
         line = "-" * 80
+
+        # Current time, formatted
+        now_str = datetime.now().strftime(r"%H:%M:%S")
+
+        # Save and print time elapsed since last epoch
+        now = time.time()
+        elapsed = now - getattr(self, "_elapsed", now)
+        setattr(self, "_elapsed", now)
+
+        if elapsed == 0:
+            since_str = ""
+        else:
+            hours, remainder = divmod(int(elapsed), 3600)
+            minutes, seconds = divmod(remainder, 60)
+
+            elapsed_str = f"{minutes:02}:{seconds:02}"
+            if hours > 0:
+                elapsed_str = f"{hours:02}:" + elapsed_str
+
+            since_str = f" | Since Last {elapsed_str}"
+
         print((
             f"{head}\n"
-            f"Epoch {epoch:>4} | Population {self.setup.population}\n"
+            f"Epoch {epoch:>4} | Population {self.setup.population:>4} | "
+            f"Time {now_str}{since_str}\n"
             f"{line}\n"
             f"Scaled overall standard deviation: {es.sigma}\n"
             f"{info}\n"
@@ -1156,36 +1177,14 @@ class Access:
         pd.set_option("display.max_rows", old_max_rows)
 
 
-    def print_status_eval(self, stdout_rec, stderr_rec, crashed):
+    def print_status_eval(self, crashed):
         '''Print logged stdout and stderr messages and crashed simulations
         after evaluating an epoch.
         '''
-        line = "-" * 80
-        if len(stderr_rec):
-            stderr_rec_str = textwrap.fill(" ".join(
-                str(s) for s in stderr_rec
-            ))
-            print(
-                line + "\n" +
-                "New stderr messages were recorded while running jobs:\n" +
-                textwrap.indent(stderr_rec_str, "  ") + "\n" +
-                f"The error messages were logged in:\n  {self.paths.outputs}",
-                flush = True,
-            )
-
-        if len(stdout_rec):
-            stdout_rec_str = textwrap.fill(" ".join(
-                str(s) for s in stdout_rec
-            ))
-            print(
-                line + "\n" +
-                "New stdout messages were recorded while running jobs:\n" +
-                textwrap.indent(stdout_rec_str, "  ") + "\n" +
-                f"The output messages were logged in:\n  {self.paths.outputs}",
-                flush = True,
-            )
 
         if len(crashed):
+            line = "-" * 80
+
             crashed_str = textwrap.fill(" ".join(
                 str(c) for c in crashed
             ))
@@ -1196,13 +1195,11 @@ class Access:
                 textwrap.indent(crashed_str, "  ") + "\n" +
                 "They crashed or terminated early; for details, check the "
                 f"output logs in:\n  {self.paths.outputs}\n"
-                "The error values for these simulations were set to NaN.",
+                "The error values for these simulations were set to NaN.\n" +
+                line + "\n",
                 flush = True,
             )
 
-        if len(stderr_rec) or len(stdout_rec) or len(crashed):
-            print(line + "\n")
-
 
     def print_after_eval(
         self,
@@ -1320,6 +1317,16 @@ class Access:
             ) for i in range(pop)
         ]
 
+        output_files = [
+            open(
+                os.path.join(
+                    self.paths.outputs,
+                    f"output.{start_index + i}.log"
+                ),
+                "w",
+            ) for i in range(pop)
+        ]
+
         # Catch the KeyboardInterrupt (Ctrl-C) signal to shut down the spawned
         # processes before aborting.
         try:
@@ -1336,28 +1343,32 @@ class Access:
                 with open(parameters_paths[i], "wb") as f:
                     pickle.dump(parameters, f)
 
+                # Get job scheduling command
+                scheduler_cmd = self.scheduler.schedule(
+                    self.paths.directory,
+                    start_index + i,
+                )
+
                 processes.append(
                     subprocess.Popen(
-                        self.setup.scheduler_cmd + [
+                        scheduler_cmd + [
                             self.paths.script,
                             parameters_paths[i],
                             result_paths[i],
                         ],
-                        stdout = subprocess.PIPE,
-                        stderr = subprocess.PIPE,
-                        universal_newlines = True,      # outputs in text mode
+                        stdout = output_files[i],
+                        stderr = subprocess.STDOUT,
                     )
                 )
 
-            # Gather results
-            results, stdout_rec, stderr_rec, crashed = \
-                self.progress.gather_results(
-                    processes,
-                    self.paths,
-                    result_paths,
-                    self.multi_objective,
-                    self.verbose,
-                )
+            # Gather results and crashed simulations
+            results, crashed = self.progress.gather_results(
+                processes,
+                self.paths,
+                result_paths,
+                self.multi_objective,
+                self.verbose,
+            )
 
         except KeyboardInterrupt:
             for proc in processes:
@@ -1368,7 +1379,11 @@ class Access:
         finally:
             signal_handler.unset()
 
-        self.print_status_eval(stdout_rec, stderr_rec, crashed)
+            for of in output_files:
+                of.close()
+
+        if self.verbose >= 1:
+            self.print_status_eval(crashed)
 
         # Find number of error values returned for one successful simulation
         num_errors = 1
@@ -1752,6 +1767,159 @@ class AccessData:
         self.results_scaled = results_scaled
 
 
+    def copy(self):
+        '''Return copy of `AccessData` object.
+        '''
+        data = AccessData.empty()
+        data.paths = self.paths.copy()
+        data.parameters = self.parameters.copy()
+        data.parameters_scaled = self.parameters_scaled.copy()
+        data.scaling = self.scaling.copy()
+        data.population = self.population
+        data.num_epochs = self.num_epochs
+        data.target = self.target
+        data.seed = self.seed
+        data.epochs = self.epochs.copy()
+        data.epochs_scaled = self.epochs_scaled.copy()
+        data.results = self.results.copy()
+        data.results_scaled = self.results_scaled.copy()
+        return data
+
+
+    def save(self, dirname):
+        '''Save `AccessData` to a new directory at `dirname`.
+        '''
+        # Copy previous folder to new location
+        shutil.copytree(self.paths.directory, dirname)
+
+        self.paths.update_paths(dirname)
+
+        # Save history
+        to_pad = len(self.results.columns) - len(self.parameters) - 1
+        columns = self.parameters.index.to_list() + [
+            f"error{i}" for i in range(to_pad)
+        ] + ["error"]
+
+        np.savetxt(
+            self.paths.history,
+            self.results.to_numpy(),
+            header = " ".join(columns),
+        )
+
+        np.savetxt(
+            self.paths.history_scaled,
+            self.results_scaled.to_numpy(),
+            header = " ".join(columns),
+        )
+
+        # Save epochs
+        np.savetxt(
+            self.paths.epochs,
+            self.epochs.to_numpy(),
+            header = " ".join(
+                [f"{p}_mean" for p in self.parameters.index] +
+                [f"{p}_std" for p in self.parameters.index] +
+                ["overall_std"]
+            ),
+        )
+
+        np.savetxt(
+            self.paths.epochs_scaled,
+            self.epochs_scaled.to_numpy(),
+            header = " ".join(
+                [f"{p}_mean" for p in self.parameters.index] +
+                [f"{p}_std" for p in self.parameters.index] +
+                ["overall_std"]
+            ),
+        )
+
+        # Save setup
+        setup_dict = dict(
+            paths = self.paths.__dict__,
+            setup = dict(
+                parameters = self.parameters.to_dict(),
+                parameters_scaled = self.parameters_scaled.to_dict(),
+                scaling = self.scaling.tolist(),
+                population = self.population,
+                target = self.target,
+                seed = self.seed,
+            ),
+        )
+
+        with open(self.paths.setup, "w") as f:
+            toml.dump(setup_dict, f)
+
+
+    def __getitem__(self, index):
+        # Select AccessData epochs
+        if isinstance(index, int):
+            # Allow negative indices
+            while index < 0:
+                index += self.num_epochs
+
+            if index >= self.num_epochs:
+                raise IndexError(textwrap.fill((
+                    f"The index=`{index}` is out of bounds for AccessData "
+                    f"with {self.num_epochs} epochs."
+                )))
+
+            data = self.copy()
+            data.num_epochs = 1
+            data.epochs = self.epochs.iloc[index:index + 1]
+            data.epochs_scaled = self.epochs_scaled.iloc[index:index + 1]
+            data.results = self.results.iloc[
+                index * self.population:(index + 1) * self.population
+            ]
+            data.results_scaled = self.results_scaled.iloc[
+                index * self.population:(index + 1) * self.population
+            ]
+            return data
+
+        elif isinstance(index, slice):
+            if index.step is not None and index.step != 1:
+                raise ValueError(textwrap.fill((
+                    "Indexing with a `slice.step` is not yet available. "
+                    "If this would be useful for you please get in touch."
+                )))
+
+            start = index.start if index.start is not None else 0
+            stop = index.stop if index.stop is not None else self.num_epochs
+
+            # Allow negative indices
+            while start < 0:
+                stop += self.num_epochs
+
+            while stop < 0:
+                stop += self.num_epochs
+
+            if stop > self.num_epochs or start >= self.num_epochs or \
+                    start >= stop:
+                raise IndexError(textwrap.fill((
+                    f"The slice=`{start}:{stop}` is out of bounds for "
+                    f"AccessData with {self.num_epochs} epochs."
+                )))
+
+            data = self.copy()
+            data.num_epochs = stop - start
+            data.epochs = self.epochs.iloc[start:stop]
+            data.epochs_scaled = self.epochs_scaled.iloc[start:stop]
+            data.results = self.results.iloc[
+                start * self.population:stop * self.population
+            ]
+            data.results_scaled = self.results_scaled.iloc[
+                start * self.population:stop * self.population
+            ]
+            return data
+
+        else:
+            raise TypeError(textwrap.fill((
+                "Epoch selection via subscripting is only possible with "
+                "integer / slice indices (e.g. `access_data[5]` or "
+                "`access_data[2:5]`). Received index with type "
+                f"`{type(index)}`."
+            )))
+
+
     def __repr__(self):
         name = self.__class__.__name__
         underline = "-" * 80
